// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.



#include <eopcc/cloud.hpp>
#include <eopcc/metrics.hpp>



#include <iostream>

namespace eopcc {

double Completeness(const GridModel<double>& input, const GridModel<double>& z_truth, double threshold) {
  assert(input.grid.size() == z_truth.grid.size());
  GridModel<unsigned char> valid(z_truth.extent(), z_truth.spacing());
  const int Nneig=8;
  int complete_count = 0;
  int complete_countvals=0;
  int winputgrid = input.dimX();
  int hinputgrid = input.dimY();
  const int neighposlut[8] = {-winputgrid-1,-winputgrid,-winputgrid+1,-1,+1,winputgrid-1,winputgrid,winputgrid+1};
  for (int idx=0; idx < input.grid.size(); ++idx) {
    if (input.grid[idx] != input.GetInvalidValue()) {
      if(z_truth.grid[idx] != z_truth.GetInvalidValue()) {
        complete_countvals +=1;
        if (std::abs(input.grid[idx] - z_truth.grid[idx]) < threshold)
          {
            complete_count +=1;
          }
      }
    } else {
      if(z_truth.grid[idx] != z_truth.GetInvalidValue()) {
        //avoid borders
        int py = idx/winputgrid;
        int px = idx%winputgrid;
        if(px>0 && px<winputgrid-1 && py>0 && py<hinputgrid-1) {
          //find minimum point in 8-connect nhood
          double zval = 10*threshold;
          for(int neighid=0;neighid<Nneig;neighid++) {
            int neighpos = neighposlut[neighid];
            if (input.grid[idx+neighpos] != input.GetInvalidValue())
              {
                zval = std::abs(input.grid[idx+neighpos] - z_truth.grid[idx]);
              }
            if(zval<threshold) {
              complete_count +=1;
              break;
            }
          }
        }
      }
    }
  }

  int valid_truth = 0;
  for (int idx=0; idx < z_truth.grid.size(); ++idx) {
    if (z_truth.grid[idx] != z_truth.GetInvalidValue()) {
      valid_truth += 1;
    }
  }

  //printf("%f,%f\n",((double) complete_count) / valid_truth,((double) complete_countvals) / valid_truth);
  return ((double) complete_count) / valid_truth;

	//3x3 for all points
/*	assert(input.grid.size() == z_truth.grid.size());
  GridModel<unsigned char> valid(z_truth.extent(), z_truth.spacing());
  const int Nneig=8;
  int complete_count = 0;
  int complete_countvals=0;
  int winputgrid = input.dimX();
  int hinputgrid = input.dimY();
  const int neighposlut[8] = {-winputgrid-1,-winputgrid,-winputgrid+1,-1,+1,winputgrid-1,winputgrid,winputgrid+1};
  for (int idx=0; idx < input.grid.size(); ++idx) {    
      if(z_truth.grid[idx] != z_truth.GetInvalidValue()) {
        complete_countvals +=1;
        //avoid borders
        int py = idx/winputgrid;
        int px = idx%winputgrid;
        if(px>0 && px<winputgrid-1 && py>0 && py<hinputgrid-1) {
          //find minimum point in 8-connect nhood
          double zval = 10*threshold;
          for(int neighid=0;neighid<Nneig;neighid++) {
            int neighpos = neighposlut[neighid];
            if (input.grid[idx+neighpos] != input.GetInvalidValue())
              {
                zval = std::abs(input.grid[idx+neighpos] - z_truth.grid[idx]);
              }
            if(zval<threshold) {
              complete_count +=1;
              break;
			}
          }
        }
		else//in border use current point
		{
			if ( input.grid[idx] != input.GetInvalidValue() &&
				std::abs(input.grid[idx] - z_truth.grid[idx]) < threshold)
	        {
		        complete_count +=1;
			}
		}

    }
  }

  int valid_truth = 0;
  for (int idx=0; idx < z_truth.grid.size(); ++idx) {
    if (z_truth.grid[idx] != z_truth.GetInvalidValue()) {
      valid_truth += 1;
    }
  }

  //printf("%f,%f\n",((double) complete_count) / valid_truth,((double) complete_countvals) / valid_truth);
  return ((double) complete_count) / valid_truth;*/

}


double ZMeanError(const GridModel<double>& input, const GridModel<double>& z_truth) {
  assert(input.grid.size() == z_truth.grid.size());

  double error = 0.0;
  long long count = 0;

  double invalid_input = input.GetInvalidValue();
  double invalid_truth = z_truth.GetInvalidValue();

  for (int idx=0; idx < input.grid.size(); idx++) {
    double input_val = input.grid[idx];
    double truth_val = z_truth.grid[idx];

    if (input_val != invalid_input && invalid_truth != truth_val) {
      error += input_val - truth_val;
      count += 1;
    }
  }

  return error / count;
}


double ZRMSError(const GridModel<double>& input, const GridModel<double>& z_truth, double zshift) {
  assert(input.grid.size() == z_truth.grid.size());

  double error = 0.0;
  long long count = 0;

  double invalid_input = input.GetInvalidValue();
  double invalid_truth = z_truth.GetInvalidValue();

  for (int idx=0; idx < input.grid.size(); idx++) {
    double input_val = input.grid[idx];
    double truth_val = z_truth.grid[idx];

    if (input_val != invalid_input && invalid_truth != truth_val) {
      error += pow(input_val - truth_val + zshift, 2);
      count += 1;
    }
  }
  return std::sqrt(error / count);

  //with 3x3 for all points except borders
  /*assert(input.grid.size() == z_truth.grid.size());

  double error = 0.0;
  long long count = 0;
  const int Nneig=8;
  int winputgrid = input.dimX();
  int hinputgrid = input.dimY();
  const int neighposlut[8] = {-winputgrid-1,-winputgrid,-winputgrid+1,-1,+1,winputgrid-1,winputgrid,winputgrid+1};
  double invalid_input = input.GetInvalidValue();
  double invalid_truth = z_truth.GetInvalidValue();
  double minval = 1000;
  for (int idx=0; idx < input.grid.size(); idx++) {
    double input_val = input.grid[idx];
    double truth_val = z_truth.grid[idx];	
    if(truth_val != z_truth.GetInvalidValue()) {
        //avoid borders
        int py = idx/winputgrid;
        int px = idx%winputgrid;
        if(px>0 && px<winputgrid-1 && py>0 && py<hinputgrid-1) {
          //find minimum point in 8-connect nhood 
			minval = 1000;
			double zval = minval;
			bool neighavail = false;
            for(int neighid=0;neighid<Nneig;neighid++) {
				int neighpos = neighposlut[neighid];
				if (input.grid[idx+neighpos] != input.GetInvalidValue())
				{
					zval = std::abs(input.grid[idx+neighpos] - truth_val + zshift);				  
					if(!neighavail)
						minval = zval; //initialization with a real value
					if(zval<minval) 
						minval=zval;					
					neighavail = true;
				}
			}
			if(neighavail)
			{
				error += pow(minval, 2);
				count += 1;
			}        
		}
		else//in border use current point
		{
			if (input.grid[idx] != input.GetInvalidValue())
			{
				double zval = std::abs(input.grid[idx] - truth_val + zshift);
				error += pow(zval, 2);
				count += 1;
			}

		}
	}
	
	}	
  return std::sqrt(error / count);*/
}

double ZMedianError(const GridModel<double>& input, const GridModel<double>& z_truth,
                    std::vector<double>& median_buffer, double zshift, bool use_abs) {
  double error = 0.0;
  long long count = 0;

  double invalid_input = input.GetInvalidValue();
  double invalid_truth = z_truth.GetInvalidValue();
  int cnt_invalid=0;//mbr
  int buf_idx = 0;

  for (int idx=0; idx < input.grid.size(); idx++) {
	  double input_val = input.grid[idx];
	  double truth_val = z_truth.grid[idx];

    if (input_val != invalid_input && invalid_truth != truth_val) {
      double error = (input_val - truth_val + zshift);
      if (use_abs) {
	      error = std::abs(error);
      }

      median_buffer[buf_idx] = error;
      buf_idx++;
    }
  }

  std::nth_element(median_buffer.begin(), median_buffer.begin() + buf_idx/2, median_buffer.begin() + buf_idx);

  return median_buffer[buf_idx/2];
}

std::vector<double> ZCumulativeHistogram(const GridModel<double>& input, const GridModel<double>& z_truth,
                      const std::vector<double>& errorTHs, double zshift) {
  //code for cumulative error histogram
  double error = 0.0;
  std::vector<long long> counthist(5,0);
  std::vector<double> cumul_hist(5,0.0);
  double count = 0;

  double invalid_input = input.GetInvalidValue();
  double invalid_truth = z_truth.GetInvalidValue();

  int buf_idx = 0;
  double cnt_invalid = 0;
  for (int idx=0; idx < input.grid.size(); idx++) {
    double input_val = input.grid[idx];
    double truth_val = z_truth.grid[idx];

    if (input_val != invalid_input && invalid_truth != truth_val)
    {
      double error = std::abs(input_val - truth_val + zshift);
      /*if(error!=0)
        printf("11 %f %f %f\n",input_val,truth_val,count);*/
      count++;
      if(error>errorTHs[0])
        counthist[0]++;
      if(error>errorTHs[1])
        counthist[1]++;
      if(error>errorTHs[2])
        counthist[2]++;
      if(error>errorTHs[3])
        counthist[3]++;
      if(error>= errorTHs[4])
        counthist[4]++;
      }

  }
  for (int idx=0; idx < z_truth.grid.size(); ++idx) {
    if (z_truth.grid[idx] != z_truth.GetInvalidValue()) {
    cnt_invalid += 1;
    }
  }
  count = (count==0) ? 1 : count;
  cnt_invalid = (cnt_invalid==0) ? 1 : cnt_invalid;
  cumul_hist[0] = counthist[0]/count;
  cumul_hist[1] = counthist[1]/count;
  cumul_hist[2] = counthist[2]/count;
  cumul_hist[3] = counthist[3]/count;
  cumul_hist[4] = counthist[4]/count;

  return cumul_hist;
}

bool ascendingSort(double i, double j) { return i < j; }

std::vector<double> ZErrorPercentile(const GridModel<double>& input, const GridModel<double>& z_truth,
                                     const std::vector<double>& errorPercentile, double zshift) {
  //code for cumulative error histogram
  double count = 0;
  double cnt_invalid = 0;
  std::vector<double> buffer(z_truth.grid.size());
  std::vector<double> error_percentile(6,0.0);

  //obtain number of valid points (in range) and accumulate error buffer
  double invalid_input = input.GetInvalidValue();
  double invalid_truth = z_truth.GetInvalidValue();

  for (int idx=0; idx < input.grid.size(); idx++)
    {
      double input_val = input.grid[idx];
    double truth_val = z_truth.grid[idx];
    if (input_val != invalid_input && invalid_truth != truth_val)
    {
      double error = std::abs(input_val - truth_val + zshift);
        buffer[count] = error;
      count++;
    }

  }

  for (int idx=0; idx < z_truth.grid.size(); ++idx) {
    if (z_truth.grid[idx] != z_truth.GetInvalidValue()) {
      cnt_invalid += 1;
    }
  }

  std::sort(buffer.begin(), buffer.begin()+count,ascendingSort);

  for (int i=0;i<errorPercentile.size();i++)
  {
    double index = floor(count*errorPercentile[i]);
    if (count*errorPercentile[i]>count)
      error_percentile[i] = -1;
    else
      error_percentile[i] = buffer[(int)index];

  }

  return error_percentile;
}

GridModel<double> ZRMSErrorGrid(const GridModel<double>& input, const GridModel<double>& z_truth, double zshift) {
  GridModel<double> error_rms(input.extent(), input.spacing());
  error_rms.setAll(0.0);
  error_rms.SetWKT(input.GetWKT());

  double error = 0.0;
  long long count = 0;

  double invalid_input = input.GetInvalidValue();
  double invalid_truth = z_truth.GetInvalidValue();

  for (int idx=0; idx < input.grid.size(); idx++) {
    double input_val = input.grid[idx];
    double truth_val = z_truth.grid[idx];

    if (input_val != invalid_input && invalid_truth != truth_val) {
      double error = pow(input_val - truth_val + zshift, 2);
      error_rms.grid[idx] = error;
    }
  }

  return error_rms;
}

} // namespace eopcc
